@@ -1,18 +1,23 @@
 """Missing batteries from `torch.utils.data`."""
 
-__all__ = ['Enumerate', 'collate', 'concat', 'iloader', 'iter_batches']
-
 import itertools
-from typing import Any, Dict, Iterable, Iterator, Tuple, TypeVar, Union
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Iterator,
+    Optional,
+    Tuple,
+    TypeVar,
+    Union,
+)
 
 import numpy as np
 import torch
 from torch.utils.data import DataLoader, Dataset, TensorDataset
 
-from ._util import is_namedtuple
-
 T = TypeVar('T')
-S = TypeVar('S')
 
 
 class Enumerate(Dataset):
@@ -23,7 +28,8 @@ class Enumerate(Dataset):
     .. testcode::
 
         from torch.utils.data import DataLoader, TensorDataset
-        dataset = TensorDataset(torch.randn(9, 2), torch.randn(9))  # X, y
+        X, y = torch.randn(9, 2), torch.randn(9)
+        dataset = TensorDataset(X, y)
         for batch_idx, batch in DataLoader(Enumerate(dataset), batch_size=3):
             print(batch_idx)
 
@@ -35,6 +41,11 @@ class Enumerate(Dataset):
     """
 
     def __init__(self, dataset: Dataset) -> None:
+        """Initialize self.
+
+        Args:
+            dataset
+        """
         self._dataset = dataset
 
     @property
@@ -61,6 +72,166 @@ class Enumerate(Dataset):
         return index, self._dataset[index]
 
 
+class FnDataset(Dataset):
+    """A thin wrapper around a loader function and its arguments.
+
+    `FnDataset` allows to avoid implementing Dataset-classes (well, at least in simple
+    cases). Below you can find the full tutorial and typical use cases, but here is a
+    quick example:
+
+    Without `FnDataset`::
+
+        class ImagesList(Dataset):
+            def __init__(self, filenames, transform):
+                self.filenames = filenames
+                self.transform = transform
+
+            def __len__(self):
+                return len(self.filenames)
+
+            def __getitem__(self, index):
+                return self.transform(Image.open(self.filenames[index]))
+
+        dataset = ImagesList(filenames, transform)
+
+    With `FnDataset`::
+
+        dataset = FnDataset(Image.open, filenames, transform)
+
+    .. rubric:: Tutorial
+
+    With vanilla PyTorch, in order to create a dataset you have to inherit from
+    `torch.utils.data.Dataset` and implement three methods:
+
+    - :code:`__init__`
+    - :code:`__len__`
+    - :code:`__getitem__`
+
+    With `FnDataset` the only thing you *may need* to implement is the :code:`fn`
+    argument that will power :code:`__getitem__`. The easiest way to learn
+    `FnDataset` is to go through examples below.
+
+    A list of images::
+
+        dataset = FnDataset(Image.open, filenames)
+        # dataset[i] returns Image.open(filenames[i])
+
+    A list of images that are cached after the first load::
+
+        from functools import lru_cache
+        dataset = FnDataset(lru_cache(None)(Image.open), filenames)
+
+    `pathlib.Path` is very useful when you want to create a dataset that reads from
+    files. For example::
+
+        images_dir = Path(...)
+        dataset = FnDataset(Image.open, images_dir.iterdir())
+
+    If there are many files, but you need only those with specific extensions, use
+    `pathlib.Path.glob`::
+
+        dataset = FnDataset(Image.open, images_dir.glob(*.png))
+
+    If there are many files in many subfolders, but you need only those with specific
+    extensions and that satisfy some condition, use `pathlib.Path.rglob`::
+
+        dataset = FnDataset(
+            Image.open, (x for x in images_dir.rglob(*.png) if condition(x))
+        )
+
+    A segmentation dataset::
+
+        image_filenames = ...
+        gt_filenames = ...
+
+        def get(i):
+            return Image.open(image_filenames[i]), Image.open(gt_filenames[i])
+
+        dataset = FnDataset(get, len(image_filenames))
+
+    A dummy dataset that demonstrates that `FnDataset` is a very general thing:
+
+    .. testcode::
+
+        def f(x):
+            return x * 10
+
+        def g(x):
+            return x * 2
+
+        dataset = FnDataset(f, 3, g)
+        # dataset[i] returns g(f(i))
+        assert len(dataset) == 3
+        assert dataset[0] == 0
+        assert dataset[1] == 20
+        assert dataset[2] == 40
+
+    """
+
+    def __init__(
+        self,
+        fn: Callable[..., T],
+        args: Union[int, Iterable],
+        transform: Optional[Callable[[T], Any]] = None,
+    ) -> None:
+        """Initialize self.
+
+        Args:
+            fn: the function that produces values based on arguments from :code:`args`
+            args: arguments for :code:`fn`. If an iterable, but not a list, then is
+                casted to a list. If an integer, then the behavior is the same as for
+                :code:`list(range(args))`. The size of :code:`args` defines the return
+                value for `FnDataset.__len__`.
+            transform: if presented, is applied to the return value of `fn` in
+                `FnDataset.__getitem__`
+
+        Examples:
+            .. code-block::
+
+                import PIL.Image as Image
+                import torchvision.transforms as T
+
+                dataset = FnDataset(Image.open, filenames, T.ToTensor())
+        """
+        self._fn = fn
+        if isinstance(args, Iterable):
+            if not isinstance(args, list):
+                args = list(args)
+        self._args = args
+        self._transform = transform
+
+    def __len__(self) -> int:
+        """Get the dataset size.
+
+        See `FnDataset` for details.
+
+        Returns:
+            size
+        """
+        return len(self._args) if isinstance(self._args, list) else self._args
+
+    def __getitem__(self, index: int) -> Any:
+        """Get value by index.
+
+        See `FnDataset` for details.
+
+        Args:
+            index
+        Returns:
+            value
+        Raises:
+            IndexError: if :code:`index >= len(self)`
+        """
+        if isinstance(self._args, list):
+            x = self._args[index]
+        elif index < self._args:
+            x = index
+        else:
+            raise IndexError(f'Index {index} is out of range')
+        x = self._fn(x)
+        return x if self._transform is None else self._transform(x)
+
+
 class _IndicesDataset(Dataset):
     def __init__(self, size: int) -> None:
         self.size = size
@@ -72,29 +243,20 @@ class _IndicesDataset(Dataset):
         return i
 
 
-def iloader(size: int, *args, **kwargs) -> DataLoader:
-    """Make `~torch.utils.data.DataLoader` over batches of indices.
+class IndexLoader:
+    """Like `~torch.utils.data.DataLoader`, but over indices instead of data.
 
-    **The shuffling logic is fully delegated to native PyTorch DataLoader**, i.e. no
-    custom logic is performed under the hood.
-
-    Args:
-        size: the size of dataset (for example, :code:`len(dataset)`)
-        *args: positional arguments for `torch.utils.data.DataLoader`
-        **kwargs: keyword arguments for `torch.utils.data.DataLoader`
-    Raises:
-        AssertionError: if size is not positive
-
-    See Also:
-        `iter_batches`
+    **The shuffling logic is delegated to the native PyTorch DataLoader**, i.e. no
+    custom logic is performed under the hood. The data loader which actually generates
+    indices is available as `IndexLoader.loader`.
 
     Examples:
         Usage for training:
 
         .. code-block::
 
-            train_loader = iloader(len(train_dataset), batch_size, shuffle=True)
-            for epoch in epoches:
+            train_loader = IndexLoader(len(train_dataset), batch_size, shuffle=True)
+            for epoch in epochs:
                 for batch_idx in train_loader:
                     ...
 
@@ -103,7 +265,7 @@ def iloader(size: int, *args, **kwargs) -> DataLoader:
         .. testcode::
 
             dataset_size = 10  # len(dataset)
-            for batch_idx in iloader(dataset_size, batch_size=3):
+            for batch_idx in IndexLoader(dataset_size, batch_size=3):
                 print(batch_idx)
 
         .. testoutput::
@@ -116,7 +278,7 @@ def iloader(size: int, *args, **kwargs) -> DataLoader:
         .. testcode::
 
             dataset_size = 10  # len(dataset)
-            for batch_idx in iloader(dataset_size, 3, drop_last=True):
+            for batch_idx in IndexLoader(dataset_size, 3, drop_last=True):
                 print(batch_idx)
 
         .. testoutput::
@@ -124,33 +286,83 @@ def iloader(size: int, *args, **kwargs) -> DataLoader:
             tensor([0, 1, 2])
             tensor([3, 4, 5])
             tensor([6, 7, 8])
+
+    See also:
+        `zero.iter_batches`
     """
-    assert size > 0
-    return DataLoader(_IndicesDataset(size), *args, **kwargs)
+
+    def __init__(
+        self, size: int, *args, device: Union[int, str, torch.device] = 'cpu', **kwargs
+    ) -> None:
+        """Initialize self.
+
+        Args:
+            size: the number of items (for example, :code:`len(dataset)`)
+            *args: positional arguments for `torch.utils.data.DataLoader`
+            device: if not CPU, then all indices are materialized and moved to the
+                device at the beginning of every loop. It can be useful when the indices
+                are applied to non-CPU data (e.g. CUDA-tensors) and moving data between
+                devices takes non-negligable time (which can happen in the case of
+                simple and fast models like MLPs).
+            **kwargs: keyword arguments for `torch.utils.data.DataLoader`
+        Raises:
+            AssertionError: if size is not positive
+        """
+        assert size > 0
+        self._batch_size = args[0] if args else kwargs.get('batch_size', 1)
+        self._loader = DataLoader(_IndicesDataset(size), *args, **kwargs)
+        if isinstance(device, (int, str)):
+            device = torch.device(device)
+        self._device = device
+
+    @property
+    def loader(self) -> DataLoader:
+        """The underlying DataLoader."""
+        return self._loader
+
+    def __len__(self) -> int:
+        """Get the size of the underlying DataLoader."""
+        return len(self.loader)
+
+    def __iter__(self):
+        return iter(
+            self._loader
+            if self._device.type == 'cpu'
+            else torch.cat(list(self.loader)).to(self._device).split(self._batch_size)
+        )
+
+
+def _is_namedtuple(x) -> bool:
+    return isinstance(x, tuple) and all(
+        hasattr(x, attr) for attr in ['_make', '_asdict', '_replace', '_fields']
+    )
 
 
 def iter_batches(
     data: Union[
-        torch.Tensor, Tuple[torch.Tensor, ...], Dict[Any, torch.Tensor], TensorDataset,
+        torch.Tensor,
+        Tuple[torch.Tensor, ...],
+        Dict[Any, torch.Tensor],
+        TensorDataset,
     ],
     *args,
     **kwargs,
 ) -> Iterator:
-    """*Efficiently* iterate over data in a batchwise manner.
+    """*Efficiently* iterate over data (tensor, tuple of tensors, dict of tensors etc.)
+    in a batchwise manner.
 
     The function is useful when you want to *efficiently* iterate **once** over
     tensor-based data in a batchwise manner. See examples below for typical use cases.
 
     The function is a more efficient alternative to `torch.utils.data.DataLoader` when
     it comes to in-memory data, because it uses batch-based indexing instead of
-    item-based indexing (DataLoader's behavior). **The shuffling logic is fully
-    delegated to native PyTorch DataLoader**, i.e. no custom logic is performed under
-    the hood.
+    item-based indexing (DataLoader's behavior). **The shuffling logic is delegated to
+    the native PyTorch DataLoader**, i.e. no custom logic is performed under the hood.
 
     Args:
         data:
-        *args: positional arguments for `iloader`
-        **kwargs: keyword arguments for `iloader`
+        *args: positional arguments for `IndexLoader`
+        **kwargs: keyword arguments for `IndexLoader`
     Returns:
         Iterator over batches.
 
@@ -163,8 +375,8 @@ def iter_batches(
         If you want to infititely iterate over batches, wrap the function in
         :code:`while True:`.
 
-    See Also:
-        - `iloader`
+    See also:
+        - `zero.data.IndexLoader`
         - `concat`
 
     Examples:
@@ -179,37 +391,43 @@ def iter_batches(
 
         .. code-block::
 
-            for epoch in epoches:
+            for epoch in epochs:
                 for batch in iter_batches(data, batch_size, shuffle=True)):
                     ...
     """
     # mypy understands very little about this function
-    if is_namedtuple(data):
+    if _is_namedtuple(data):
         assert data
         f = lambda idx: type(data)._make(x[idx] for x in data)  # type: ignore # noqa
-        size = len(data[0])
+        size = len(data[0])  # type: ignore # noqa
     elif isinstance(data, tuple):
         assert data
         f = lambda idx: type(data)(x[idx] for x in data)  # type: ignore # noqa
-        size = len(data[0])
+        size = len(data[0])  # type: ignore # noqa
     elif isinstance(data, dict):
         assert data
         f = lambda idx: type(data)({k: v[idx] for k, v in data.items()})  # type: ignore # noqa
-        size = len(next(iter(data.values())))
+        size = len(next(iter(data.values())))  # type: ignore # noqa
     else:
         f = data.__getitem__
-        size = len(data)
-    return map(f, iloader(size, *args, **kwargs))
+        size = len(data)  # type: ignore # noqa
+    return map(f, IndexLoader(size, *args, **kwargs))
 
 
-def concat(iterable: Iterable[T]) -> Union[S, Tuple[S, ...], Dict[Any, S]]:
-    """Concatenate items of the iterable along the first dimension.
+def concat(iterable: Iterable[T]) -> T:
+    """Concatenate items (tensors, numpy-arrays, tuples, dicts etc.) along the first
+    dimension.
 
-    Use intuition, the tutorial and examples (see below) to understand what the function
-    does. Technical specification is unlikely to help here, so there is none :)
+    `concat` is a more general version of :code:`torch.cat(..., dim=0)`. It works not
+    only with sequences of tensors, but also with sequences of containers (tuples,
+    dicts etc.) of different types of data (tensors, numpy-arrays, primitive types). See
+    the tutorial and the examples below to understand what the function does.
 
     Args:
-        iterable: items **of the same structure**
+        iterable: items **of the same structure** (for example, "an iterable of tensors"
+            OR "an iterable of tuples of tensors where all the tuples are of the same
+            length" OR "an iterable of dicts of tensors and numpy-arrays where all the
+            dicts have the same keys" etc.)
     Returns:
         Concatenated items of the iterable.
 
@@ -222,10 +440,10 @@ def concat(iterable: Iterable[T]) -> Union[S, Tuple[S, ...], Dict[Any, S]]:
     Warning:
         The function starts with conversion of the iterable to a list. Make sure that
         you have enough memory for such operation, otherwise, memory limit may be
-        exceeded. Note that manual implementation would involve the same conversion,
-        just keep this in mind when using the function.
+        exceeded. Note that in most cases manual implementation would involve the same
+        conversion, just keep this in mind when using the function.
 
-    See Also:
+    See also:
         `iter_batches`
 
     .. rubric:: Tutorial
@@ -415,7 +633,7 @@ def concat(iterable: Iterable[T]) -> Union[S, Tuple[S, ...], Dict[Any, S]]:
     first = data[0]
     return (
         type(first)._make(concat_fn(x[i] for x in data) for i, _ in enumerate(first))
-        if is_namedtuple(first)
+        if _is_namedtuple(first)
         else type(first)(concat_fn(x[i] for x in data) for i, _ in enumerate(first))
         if isinstance(first, tuple)
         else type(first)((key, concat_fn(x[key] for x in data)) for key in first)
@@ -433,4 +651,5 @@ def collate(iterable: Iterable[T]) -> Any:
     """
     if not isinstance(iterable, list):
         iterable = list(iterable)
-    return torch.utils.data.dataloader.default_collate(iterable)
+    # > Module has no attribute "default_collate"
+    return torch.utils.data.dataloader.default_collate(iterable)  # type: ignore
