@@ -3,13 +3,11 @@ from collections import namedtuple
 import pytest
 import torch
 
-from zero.metrics import Metric
-
 Point = namedtuple('Point', ['x', 'y'])
 Model = namedtuple('Model', ['model', 'weight', 'bias', 'loss_fn', 'optimizer', 'data'])
 
 
-class ObjectCounter(Metric):
+class ObjectCounter:
     def __init__(self, sign):
         self.sign = sign
         self.reset()
